"use strict";
this.name        = "fighters-ship";
this.author      = "Norby";
this.copyright   = "2016 Norbert Nagy";
this.licence     = "CC BY-NC-SA 4.0";
this.description = "ship script of fighters";

this.$FighterKey = ""; //store equipmentKey of this fighter to award back when landed
this.$FighterHull = this.ship.maxEnergy; //must repair in hangar, avoid recharge when no shields
this.$FighterNumber = 0; //order of this fighter in escort formation
this.$FighterOwner = null; //store the mothership
this.$FighterShields = true; //cache of this.ship.scriptInfo.npc_shields
this.$ForcedLanding = false; //forced due to damage, do not stop landing with others
this.$OwnerScoopPos = null; //store the mothership's scoop position for landing
this.$SniperLock = false; //cache of shipdata.plist script_info fighters-sniperlock field
this.$TimerLandNow = null; //trigger _landNow() after some time for sure landing
this.$ws = worldScripts.fighters;

//ship script events
this.shipAttackedOther = function(other) {
        if( other && other.isDerelict && other == this.ship.target )
                this.ship.target = null; //do not destroy derelicts, choose another target instead
}

this.shipBountyChanged = function(delta,reason) {
        //prevent orange scanner flag on player's fighters with ShipVersion OXP
        if(this.$FighterOwner == player.ship && this.ship.bounty > 0 ) {
                this.ship.bounty = 0; //Warning: this change trigger this event again!
        }                             //So must check before that bounty is greater than 0
}                                     //else this is an infinite loop which make Crash To Desktop!

this.shipCollided = this.shipWasScooped = function(otherShip) {
        if(otherShip == this.$FighterOwner) {
            this.ship.energy = this.ship.maxEnergy; //sometimes prevent destruction

            if(this.ship.AIState == "LANDING") //put into the hangar
                this._landNow(this.ship, this.$FighterOwner);
        }
}

this.shipLaunchedEscapePod = function(escapepod, passengers) {
        this.ship.target = null;
        this.ship.performHold();
        this.ship.switchAI("nullAI.plist");
}

this.shipTakingDamage = function(amount, whom, type) {
        this.$FighterHull -= amount; //workaround recharge when no shields
        if( !this.$FighterShields ) { //light fighter without shields
                if(this.$FighterHull < 0) { //out of hull
                        delete this.shipTakingDamage;//prevent infinite loop
                        if(this.ship && this.ship.isValid) {
                                this.ship.abandonShip(); //eject pilot
                                this.ship.explode(); //destroy the defeated fighter
                        }
                } else if(this.$FighterHull < this.ship.maxEnergy * 0.4) { //damaged
			var fi = this.ship;
                        var fai = "fightersAI.plist";
                        if(fi.AI != fai ) fi.switchAI(fai);
                        fi.target = null; //clear target to stop attack
                        fi.AIState = "LANDING";
                        this.$ForcedLanding = true;
                }
        } else { //make sure damages on heavy fighters are persistent by set ECM to damaged
                if(this.ship && this.ship.energy < 0.4 * this.ship.maxEnergy ) {
                        if( this.ship.equipmentStatus("EQ_ECM") == "EQUIPMENT_OK" ) {
                                this.ship.setEquipmentStatus("EQ_ECM", "EQUIPMENT_DAMAGED");
                                if(this.ship.dataKey === "fighters-raider") { //lose a laser
                                        var su = this.ship.subEntities;
                                        if(su && su[1]) su[1].remove();//no true within to display message
                                }
                        }
                }
        }
}

this.shipSpawned = function() {
        if( !this.ship ) return;
        var si = this.ship.scriptInfo;
        if( si && si.npc_shields == "no" //Swarm has regenerating armor without CustiomShields
            && this.ship.primaryRole != "EQ_FIGHTER_SWARM" ) {
                this.$FighterShields = false;
        } else this.$FighterShields = true;
        
        if( si && si.fighters_sniperlock ) this.$SniperLock = true;
        
        if( this.ship.primaryRole === "EQ_FIGHTER_RAIDER" //stealth fighters in cloaked era
            && ( player.ship.equipmentStatus("EQ_CLOAKING_DEVICE") != "EQUIPMENT_UNAVAILABLE"
                || missionVariables.cloak === "COMPLETE"  ) ) {
            this.ship.awardEquipment("EQ_CLOAKING_DEVICE");
        }
}

this.spawnedAsEscort = function(mother) {  //for NPC mothers
        if(mother && mother.script) {
            if(!mother.script.$LaunchedFighters) mother.script.$LaunchedFighters = 1;
            else mother.script.$LaunchedFighters++;
            this.$FighterNumber = mother.script.$LaunchedFighters;
            if(!mother.script.$Fighters) mother.script.$Fighters = [];
            mother.script.$Fighters.push(this.ship);
            this.ship.displayName = this.ship.name+" #"+this.$FighterNumber;
        }
        this.$FighterOwner = mother;
}

this.shipTargetDestroyed = function(target) {
        if(this.ship == player.ship || this.$FighterOwner != player.ship )
            return; //exit if worldScript, do in ship script only
        if(target.primaryRole == "constrictor" && missionVariables.conhunt && missionVariables.conhunt == "STAGE_1") {
                // just in case an escort kills the constrictor, let's not break the mission for the player...
                missionVariables.conhunt = "CONSTRICTOR_DESTROYED";
        }
        
        if(target.isRock || target.isBoulder || target.isCargo || target.isWeapon)
                return;
        
//      player.score += 1; - score should remain personal
        var b = "";
        if(target.bounty > 0) {
            b = " : " + target.bounty + "₢ awarded.";
            player.credits += target.bounty;
        }
        player.consoleMessage(this.ship.displayName+" killed "+target.name+b, 5);
        log(this.name, this.ship.displayName+" killed " + target.name + " : " + target.bounty +"₢");
}

this.shipDied = function(whom,why) {
        if( !this.ship || this.ship == player.ship ) return; //exit if worldScript, do in ship script only
        if( this.ship.dataKey === "fighters-swarm" ) {
            var ai = system.addShips("fighters-alienitem", 1, this.ship.position, 20);
            if( ai && ai[0] && this.ship.velocity) ai[0].velocity = this.ship.velocity;
            var e = this.ship.subEntities;
            var l = 0;
            if( e ) l = e.length;
            if( l > 0 ) { //there are subents so not all swarm fighters are lost
                var sw = system.addShips("fighters-swarm", 1, this.ship.position, 20)[0];
                if(sw) {
                    sw.target = this.ship.target;
                    if(this.ship.target) sw.addDefenseTarget(this.ship.target);
                    if(sw.script) {
                        sw.script.$FighterKey = this.$FighterKey;
                        sw.script.$FighterNumber = this.$FighterNumber;
                        sw.script.$FighterOwner = this.$FighterOwner;
                        sw.script.$TrailsDisabled = 1; //reduce load with Trails OXP
                    }
                    this.$FighterOwner.escortGroup.addShip(sw);
                    var f;
                    if(this.$FighterOwner === player.ship) f = this.$ws.$Fighters;
                    else f = this.$FighterOwner.script.$Fighters;
                    var fi = f.indexOf(this.ship);
                    if( fi > -1 ) {
                        if(this.$FighterOwner === player.ship) this.$ws.$Fighters[fi] = sw;
                        else this.$FighterOwner.script.$Fighters = sw;
                    }
                    sw.velocity = this.ship.velocity;
                    sw.orientation = this.ship.orientation;
                    sw.displayName = this.ship.displayName;
                    if( this.$FighterOwner === player.ship ) {
                        sw.bounty = 0; //remove bounty from fighter given in shipdata.plist
                        sw.scannerDisplayColor1 = [0, 0.4, 0]; //green, mean frendly
                        //si.scannerDisplayColor2 = [0, 0.4, 0]; //green, mean frendly
                    } else {
                        if( ship.bounty == 0 ) sw.bounty = 0; //remove bounty from fighter given in shipdata.plist
                    }
                    var su = sw.subEntities;
                    if(su && su[1]) {
                        su[1].remove(true); //remove a swarm fighter
                        if( l == 1 ) su[0].remove(true); //remove another also
                    }
                }
            }
        }
        if( this.ship.dataKey === "fighters-shark" ) {
                var a = Math.ceil(this.$ws._alienItems("EQ_FIGHTER_SHARK")/2);
                var ai = system.addShips("fighters-alienitem", a, this.ship.position, 100);
                if( ai && this.ship.velocity )
                    for( var i = 0; i < ai.length; i++ )
                        if(ai[i]) ai[i].velocity = this.ship.velocity;
        }
        if( this.ship.dataKey === "fighters-king" ) {
                var a = Math.ceil(this.$ws._alienItems("EQ_FIGHTER_KING")/2);
                var ai = system.addShips("fighters-alienitem", a, this.ship.position, 100);
                if( ai && this.ship.velocity )
                    for( var i = 0; i < ai.length; i++ )
                        if(ai[i]) ai[i].velocity = this.ship.velocity;
        }
        if( this.$FighterOwner == player.ship ) {
                var d = " lost";
                if( whom && whom.isValid && whom != player.ship ) {
                        d = " defeated by "+whom.displayName;
                }
                player.commsMessage(this.ship.displayName + d, 10);
                this.$ws._MFD(); //update numbers in MFD
        }
        log(this.name, this.ship.displayName+" terminated by "+whom+" "+why); //log NPC fighters also
}


//fighter functions

this._escortPosition = function() {this._escortPosition2(this.ship);} //for fighterAI.plist

this._escortPosition2 = function(ship) {
        if( ship && ship.isDerelict ) {
            ship.target = null;
            ship.performHold();
            ship.switchAI("nullAI.plist");
            return;
        }
        var sc = ship.script;
        if( !sc || !sc.$FighterNumber || !sc.$FighterOwner || !sc.$FighterOwner.isValid
            || ship == player.ship ) return; //for sure

        if( sc.$FighterOwner == player.ship && //green alert or jump countdown
            ( player.alertCondition == 1 || this.$ws.$ForcedLanding ) ) {
            ship.AIState = "LANDING";
            this.$ws._MFD(); //update if owner is the player
            return;
        }

        if( ship.target && ship.target.hasHostileTarget ) {
            ship.setAI("interceptAI.plist");
            return;
        }

        var epos = sc.$ws.coordinatesForEscortPosition(sc.$FighterNumber);
        var pos = sc.$FighterOwner.position.add(epos);
        ship.savedCoordinates = pos;
        if( ship.injectorSpeedFactor > 0 )
            ship.desiredSpeed = ship.injectorSpeedFactor * ship.maxSpeed;
        else ship.desiredSpeed = 7 * ship.maxSpeed; //for Oolite 1.80
        ship.desiredRange = 0; //go to the escort position in the formation
        ship.performFlyToRangeFromDestination();
        ship.reactToAIMessage("OWNER_FAR");//go back to owner
        //var dist = ship.position.distanceTo(sc.$FighterOwner.position);
        //if(dist > 10000) ship.reactToAIMessage("OWNER_FAR");//go back to owner
        //else if(dist > 5000) ship.reactToAIMessage("OWNER_NEAR");//go back to owner
        //else ship.reactToAIMessage("OWNER_NEXT");//keep formation
}

this._landing = function() {this._landing2(this.ship);}

this._landing2 = function(ship) {
        if( ship && ship.isDerelict ) {
            ship.target = null;
            ship.performHold();
            ship.switchAI("nullAI.plist");
            return;
        }
        var sc = ship.script;
        if( !sc || ship == player.ship ) return; //for sure

        var owner = sc.$FighterOwner;
        if( !owner || !owner.isValid ) return;

        ship.target = null; //clear target to stop attack

        var landingdist = 50; //m from the scoop position of owner where put into hangar
        //make a waypoint below the owner's scoop position
        var scooppos = owner.position;
        var sp = this.$OwnerScoopPos;
        if( !sp ) { //first time read scoop pos from shipdata.plist
            var shipdata = Ship.shipDataForKey(owner.dataKey);
            if( shipdata ) {
                sp = shipdata["scoop_position"];
                if(sp && sp != "undefined") {
                    sp = sp.split(" ",3);
                    if( sp && sp.length == 3 ) {
                        this.$OwnerScoopPos = sp;
                        //log(this.name, owner.name+" scoop_position:"+sp); //debug
                    }
                }
            }
            if( !this.$OwnerScoopPos ) { //no scoop pos in shipdata so make one
                this.$OwnerScoopPos = [];
                this.$OwnerScoopPos[0] = 0;
                this.$OwnerScoopPos[1] = -20;
                this.$OwnerScoopPos[2] = 0;
                sp = this.$OwnerScoopPos;
            }
        }
        scooppos = scooppos.add(owner.vectorRight.multiply(sp[0]))
                            .add(owner.vectorUp.multiply(sp[1]))
                            .add(owner.heading.multiply(sp[2]));
        var down = 1.5*owner.collisionRadius; //AI never enter into collisionRadius
        var downpos = scooppos.add(owner.vectorUp.multiply(-down));
        ship.savedCoordinates = downpos.add(owner.velocity.multiply(6)); //go below the owner
        // also must use setDestinationFromCoordinates in AI.plist

        var d = ship.position.distanceTo(scooppos);
        var d2 = ship.position.distanceTo(downpos);
        //log(this.name, d);
        if( this.$ws.$TractorBeam ) {
                if( d < landingdist ) { //enough near?: add to the hangar
                        this._landNow(ship, owner);
                        return;
                }
                if( d2 < down ) { //arrived below the owner?
                        //if( owner.addCollisionException ) { //need Oolite v1.81
                        //    owner.addCollisionException(ship); //must go near
                            //log(this.name, "coll.ex.len:"+owner.collisionExceptions.length);//debug
                        //}

                        //go up to the owner's center point from below into the dock
                        ship.savedCoordinates = scooppos.add(owner.velocity.multiply(2));
                        // also must use setDestinationFromCoordinates in AI.plist

                        //apply "tractor beam" to pull this fighter to the scoop
                        var v = scooppos.subtract(ship.position);
                        ship.velocity = v.direction().multiply(80).add(v.multiply(0.4))
                                .add(owner.velocity);

                        //make sure the fighter will land after 10 seconds
                        if( !this.$TimerLandNow )
                                this.$TimerLandNow = new Timer(this, this._TimedLandNow.bind(this), 10);

                        ship.desiredSpeed = 0; //must set before "OWNER_NEXT"
                        ship.reactToAIMessage("OWNER_NEXT"); //go straight
                        return;
                }
        } else { //land without tractor beam
                if( d < landingdist ) {
                        //must be triggered outside collisionRadius of owner
                        this._landNow(ship, owner); //add to the hangar
                        return;
                } else if( d2 < down ) { //surely land after 10 seconds
                        if( !this.$TimerLandNow )
                                this.$TimerLandNow = new Timer(this, this._TimedLandNow.bind(this), 10);
                        //go up to the owner's center point from below into the dock
                        ship.savedCoordinates = scooppos.add(owner.velocity);
                        ship.desiredSpeed = ship.maxSpeed; //must set before "OWNER_NEXT"
                        ship.reactToAIMessage("OWNER_NEXT"); //go straight
                        return;
                }
        }

        if( d < 2000 ) ship.reactToAIMessage("OWNER_NEAR"); //approach owner at normal speed
        else ship.reactToAIMessage("OWNER_FAR"); //use injectors in approach if available
}

this._landNow = function(ship, owner) { //put back into the internal hangar
        if(!ship || !owner || !owner.isValid) return; //for sure
        var f = [], f2 = []; //remove from $Fighters array
        this.$ws._awardFighter(ship, owner); //put back into the owner's equipment list
        if( owner == player.ship ) {
            if( this.$ws.$LaunchedFighters > 0 ) this.$ws.$LaunchedFighters--;
            f = this.$ws.$Fighters;
        } else if( owner.script ) {
            if( owner.script.$LaunchedFighters > 0 ) owner.script.$LaunchedFighters--;
            f = owner.script.$Fighters;
        }
        var out = 0;
        for(var i = 0; i < f.length; i++) {
            var fi = f[i];
            if( fi != ship ) {
                f2.push(fi);
                if( fi && fi.isValid ) out++;
            }
        }
        if( owner == player.ship ) {
            this.$ws.$Fighters = f2;
            this.$ws._MFD(); //update numbers in MFD
            if( out == 0 ) {
                player.commsMessage("All fighters landed", 10);
                player.consoleMessage(" ", 10); //make room to the countdown
            }
        } else if( owner.script ) owner.script.$Fighters = f2;

        //prevent double landing if both shipCollided and shipWasScooped is fired
        var sc = ship.script;
        if( sc ) sc.$FighterOwner = null;

        //if( owner.removeCollisionException ) { //need Oolite v1.81
        //    owner.removeCollisionException(ship); //must go near
            //log(this.name, "coll.ex.len:"+owner.collisionExceptions.length);//debug
        //}
        ship.remove(true); //landed
}

this._newTarget = function() {this._newTarget2(this.ship);} //AI in plist located new target

this._newTarget2 = function() { //AI in plist located new target
        var sc = ship.script;
        if( !sc || ship == player.ship ) return; //for sure
        var owner = sc.$FighterOwner;
        if( !owner || !owner.isValid || owner != player.ship ) return; //player's fighters only
        var t = ship.target;
        if( !this.$ws._isValidFighterTarget(t, owner) ) return;

        player.consoleMessage(ship.name+" #"+this.$FighterNumber+" aim "+t.displayName, 4.5);
        this.$ws._MFD(); //update numbers in MFD
}

this._TimedLandNow = function() { 
        if( this.$TimerLandNow ) {
                this.$TimerLandNow.stop();
                delete this.$TimerLandNow;
        }
        this._landNow(this.ship, this.$FighterOwner);
}
